module Interpretter exposing (..)


import LispParser as LP exposing (AST, Statement)
import Dict exposing (Dict)
import Result


type Value
  = Symbol String
  | Integer Int
  | Quoted Value
  | Cons Value Value
  | Fun Function
  | Null


type alias Environment = Dict String Value


type Function
  = Builtin (Environment -> Value -> Result String Value)
  | Closure Environment Skeleton Value


type alias InterpreterState =
  { defs : Environment
  , macros : Environment
  , result : List (Result String Value)
  }


type Skeleton
  = ConsPattern Skeleton Skeleton
  | Capture String
  | Wildcard


interpret : InterpreterState -> List Statement -> InterpreterState
interpret state stmts = case stmts of
  [] -> state
  LP.MacroDef name ast :: rest ->
    case resolve state.macros ast |> Result.andThen (eval state.defs) of
      Err s ->
        { state
        | result =
          Err ("Error in macro definition " ++ name ++ ":\n" ++ s) ::
          state.result
        }
      Ok expr -> interpret
        { state | macros = Dict.insert name expr state.macros }
        rest
  LP.FunctionDef name ast :: rest ->
    case resolve state.macros ast |> Result.andThen (eval state.defs) of
      Err s ->
        { state
        | result =
          Err ("Error in function definition " ++ name ++ ":\n" ++ s) ::
          state.result
        }
      Ok expr -> interpret
        { state | defs = Dict.insert name expr state.defs }
        rest
  LP.RawExpression ast :: rest -> interpret
    { state
    | result = (resolve state.macros ast
                |> Result.andThen (eval state.defs)) :: state.result
    }
    rest


resolve : Environment -> AST -> Result String Value
resolve macros ast = case ast of
  LP.Variable s -> Ok (Symbol s)
  LP.Integer n -> Ok (Integer n)
  LP.Quoted q -> resolve macros q |> Result.map Quoted
  LP.ConsList (LP.Variable name :: rest) -> case resolveList macros rest of
    Ok args -> case Dict.get name macros of
      Nothing -> Ok (Cons (Symbol name) args)
      Just macro -> call Dict.empty macro args
    Err s -> Err s
  LP.ConsList rest -> resolveList macros rest


resolveList : Environment -> List AST -> Result String Value
resolveList macros l = case l of
  head :: rest -> case resolve macros head of
    Err s -> Err s
    Ok headValue -> case resolveList macros rest of
      Err s -> Err s
      Ok restValue -> Ok (Cons headValue restValue)
  [] -> Ok Null


eval : Environment -> Value -> Result String Value
eval defs val = case val of
  Symbol s -> case Dict.get s defs of
    Just v -> Ok v
    Nothing -> Err ("Unknown symbol: " ++ s)
  Quoted expr -> Ok expr
  Cons carExpr cdrExpr -> case eval defs carExpr of
    Err s -> Err s
    Ok car -> case calllessEval defs cdrExpr of
      Err s -> Err s
      Ok cdr -> call defs car cdr
  _ -> Ok val


calllessEval : Environment -> Value -> Result String Value
calllessEval defs val = case val of
  Cons carExpr cdrExpr -> case eval defs carExpr of
    Err s -> Err s
    Ok car -> case calllessEval defs cdrExpr of
      Err s -> Err s
      Ok cdr -> Ok (Cons car cdr)
  _ -> eval defs val


call : Environment -> Value -> Value -> Result String Value
call defs fun args = case fun of
  Fun func -> case func of
    Builtin f -> f defs args
    Closure env params body -> case assign env params args of
      Just x -> eval x body
      Nothing ->
        Err ("Bad argument shape! Expected:\n" ++ skeletonToString params
             ++ "\nGot:\n" ++ valueToString body ++ "\n")
  _ -> Err ("Attempt to call a value that is not callable:\n"
            ++ valueToString fun)


assign : Environment -> Skeleton -> Value -> Maybe Environment
assign env pattern args = case pattern of
  Wildcard -> Just env
  Capture s -> Just (Dict.insert s args env)
  ConsPattern carPattern cdrPattern ->
    case args of
      Cons car cdr -> assign env carPattern car
        |> Maybe.andThen (\x -> assign x cdrPattern cdr)
      _ -> Nothing


skeletonToString : Skeleton -> String
skeletonToString s =
  case toConsListPattern s of
    Just l -> "(" ++ consListPatternToString l ++ ")"
    Nothing -> case s of
      Wildcard -> "()"
      Capture name -> name
      ConsPattern car cdr ->
        "(. " ++ skeletonToString car ++ " " ++ skeletonToString cdr ++ ")"


toConsListPattern : Skeleton -> Maybe (List Skeleton)
toConsListPattern s = case s of
  Wildcard -> Just []
  ConsPattern car cdr -> case toConsListPattern cdr of
    Nothing -> Nothing
    Just cdrList -> Just (car :: cdrList)
  _ -> Nothing


consListPatternToString : List Skeleton -> String
consListPatternToString l = case l of
  head :: [] -> skeletonToString head
  head :: rest -> skeletonToString head ++ " " ++ consListPatternToString rest
  [] -> ""



valueToString : Value -> String
valueToString v =
  case toConsList v of
    Just l ->  "(" ++ consListToString l ++ ")"
    Nothing -> case v of
      Null -> "()"
      Symbol s -> s
      Cons car cdr ->
        "(. " ++ valueToString car ++ " " ++ valueToString cdr ++ ")"
      Quoted e -> "'" ++ valueToString e
      Integer n -> String.fromInt n
      Fun (Builtin _) -> "~builtin~"
      Fun (Closure _ params _) ->
        "(~closure~ " ++ skeletonToString params ++ ")"


toConsList : Value -> Maybe (List Value)
toConsList s = case s of
  Null -> Just []
  Cons car cdr -> case toConsList cdr of
    Nothing -> Nothing
    Just cdrList -> Just (car :: cdrList)
  _ -> Nothing


consListToString : List Value -> String
consListToString l = case l of
  head :: [] -> valueToString head
  head :: rest -> valueToString head ++ " " ++ consListToString rest
  [] -> ""
