module LispParser exposing (..)


import Lexer as L exposing (Token)


-- Syntax:
-- Program    := { Statement }
-- Statement  := MacroDef | FunDef | Expression
-- MacroDef   := 'macro:' Identifier Expression
-- FunDef     := 'fun:' Identifier Expression
-- Expression := Identifier | Number | '\'' Expression | ConsExpr
-- ConsExpr   := '(' { Expression } ')'


type AST
  = Variable String
  | Integer Int
  | Quoted AST
  | ConsList (List AST)


type Statement
  = MacroDef String AST
  | FunctionDef String AST
  | RawExpression AST


parse : List Token -> Result String (List Statement)
parse t = case t of
  [] -> Ok []
  L.Identifier "macro:" :: L.Identifier name :: tokens ->
    case parseExpr tokens of
      Ok (macro, ts) -> case parse ts of
        Ok stmts -> MacroDef name macro :: stmts |> Ok
        Err s -> Err s
      Err s -> Err s
  L.Identifier "fun:" :: L.Identifier name :: tokens ->
    case parseExpr tokens of
      Ok (function, ts) -> case parse ts of
        Ok stmts -> FunctionDef name function :: stmts |> Ok
        Err s -> Err s
      Err s -> Err s
  _ ->
    case parseExpr t of
      Ok (expr, ts) -> case parse ts of
        Ok stmts -> RawExpression expr :: stmts |> Ok
        Err s -> Err s
      Err s -> Err s


parseExpr : List Token -> Result String (AST, List Token)
parseExpr t = case t of
  L.Number x :: ts -> Ok (Integer x, ts)
  L.Identifier s :: ts -> Ok (Variable s, ts)
  L.Quote :: ts -> case parseExpr ts of
    Ok (expr, tokens) -> Ok (Quoted expr, tokens)
    Err s -> Err s
  L.OpenParen :: ts -> case consBuilder ts of
    Ok (asts, tokens) -> Ok (ConsList asts, tokens)
    Err s -> Err s
  L.CloseParen :: ts -> Err "Unexpected ')' encountered!"
  [] -> Err "Unxpected end of text!"


consBuilder : List Token -> Result String (List AST, List Token)
consBuilder t = case t of
  L.CloseParen :: ts -> Ok ([], ts)
  [] -> Err "Unclosed '('!"
  _ -> case parseExpr t of
    Err s -> Err s
    Ok (expr, ts) -> case consBuilder ts of
      Err s -> Err s
      Ok (asts, tokens) -> Ok (expr :: asts, tokens)
