module Main exposing (..)


import Interpretter as I exposing (InterpreterState, Environment, interpret)
import Lexer exposing (lexer)
import Parser as P exposing (DeadEnd, run)
import LispParser exposing (parse)
import Dict
import Std as S exposing (stdDefs, stdMacros)
import File
import File.Select as Select
import Browser
import Browser.Events exposing (onKeyDown)
import Html exposing (Html, text, div, br, button, form, textarea, input)
import Html.Events exposing (onClick, onInput)
import Html.Attributes exposing (value, type_)
import Task exposing (perform)
import Json.Decode as Json exposing (field, string, bool)


-- Main

main = Browser.element
  { init = init
  , update = update
  , view = view
  , subscriptions = always sub
  }


init : () -> ( Model, Cmd Msg )
init _ =
  ( { defs = stdDefs
    , macros = stdMacros
    , results = []
    , input = ""
    }, Cmd.none )


-- Model

type alias Model =
  { defs : Environment
  , macros : Environment
  , results : List String
  , input : String
  }


-- Controller

type Msg
  = Run String Bool
  | UpdateInput String
  | Import
  | ReadFile File.File
  | LoadFile String String



update : Msg -> Model -> (Model, Cmd Msg)
update msg model =
  let
      resultToString r = case r of
        Ok v -> I.valueToString v
        Err s -> s
  in case msg of
    Run code control ->
      if not control || code /= "Enter" then ( model, Cmd.none ) else
        case executor
          { defs = model.defs
          , macros = model.macros
          , result = []
          } model.input of
            Err s ->
              ( { model
                | results = ("> " ++ model.input)
                  :: s
                  :: model.results
                }, Cmd.none )
            Ok state ->
              ( { model
                | defs = state.defs
                , macros = state.macros
                , results = ("> " ++ model.input) ::
                            (List.map resultToString state.result
                             ++ model.results)
                , input = ""
                }, Cmd.none )
    UpdateInput s -> ( { model | input = s }, Cmd.none )
    Import -> ( model, Select.file ["text/*"] ReadFile )
    ReadFile f -> (model, File.toString f |> perform (File.name f |> LoadFile))
    LoadFile name source -> case executor
      { defs = model.defs
      , macros = model.macros
      , result = []
      } source of
        Err s ->
          ( { model
            | results = ("> Importing " ++ name)
              :: s
              :: model.results
            }, Cmd.none )
        Ok state ->
          ( { model
            | defs = state.defs
            , macros = state.macros
            , results = ("> Importing " ++ name)
                        :: (List.map resultToString state.result
                            ++ model.results)
            }, Cmd.none )


-- Subscriptions

sub : Sub Msg
sub = onKeyDown ( Json.map2 Run (field "key" string) (field "ctrlKey" bool) )

-- Code execution

executor : InterpreterState -> String -> Result String InterpreterState
executor state source =
  run lexer source |> Result.mapError deadEndsToString
  |> Result.andThen parse |> Result.map (interpret state)


deadEndsToString : List DeadEnd -> String
deadEndsToString = List.map (\end ->
  "Lexing error on line " ++ String.fromInt end.row
  ++ " column " ++ String.fromInt end.col ++ ":\n" ++
  ( case end.problem of
    P.Expecting s -> "Expected " ++ s ++ ". This should not happen."
    P.ExpectingInt -> "Expected an integer."
    P.ExpectingHex -> "Expected a hex number. This should not happen."
    P.ExpectingOctal -> "Expected an octal number. This should not happen."
    P.ExpectingBinary -> "Expected a binary number. This should not happen."
    P.ExpectingFloat ->
      "Expected a floating point number. This should not happen."
    P.ExpectingNumber -> "Expected a number. This should not happen."
    P.ExpectingVariable -> "Expected an identifier."
    P.ExpectingSymbol s -> "Expected '" ++ s ++ "'."
    P.ExpectingKeyword s -> "Expected '" ++ s ++ "'. This should not happen."
    P.ExpectingEnd -> "Expected end of text. This should not happen."
    P.UnexpectedChar -> "Unexpected character. This should not happen."
    P.Problem s -> "Problem: " ++ s ++ ". This should not happen."
    P.BadRepeat -> "Bad repeat! This should not happen." ) )
  >> String.join "\n\n"


-- View

view : Model -> Html Msg
view model =
  div []
  [ text "Ctrl+Enter to submit. Not styled. "
  , text "For help read the file attached with the source.", br [] []
  , button [ onClick Import ] [ text "Import file" ], br [] []
  , textarea [ value model.input, onInput UpdateInput ] []
  , div [] (List.map (text) model.results |> List.intersperse (br [] []))
  ]
