module Std exposing (..)


import Interpretter as I exposing (Function, Environment, Value, Skeleton)
import Dict


add : Environment -> Value -> Result String Value
add e a =
  let
      adder args = case args of
        I.Cons (I.Integer n) b -> adder b |> Result.map ((+) n)
        I.Null -> Ok 0
        _ -> Err ("'+' expected a list of integers, got "
                  ++ I.valueToString args)
  in adder a |> Result.map I.Integer


sub : Environment -> Value -> Result String Value
sub e args = case args of
  I.Cons (I.Integer a) (I.Cons (I.Integer b) _) -> Ok (I.Integer (a - b))
  I.Cons (I.Integer a) (I.Integer b) -> Ok (I.Integer (a - b))
  I.Cons (I.Integer a) I.Null -> Ok (I.Integer (-a))
  I.Integer a -> Ok (I.Integer (-a))
  _ -> Err ("'-' expected one or two integers, got "
            ++ I.valueToString args)


mul : Environment -> Value -> Result String Value
mul e a =
  let
      muler args = case args of
        I.Cons (I.Integer n) b -> muler b |> Result.map ((*) n)
        I.Null -> Ok 1
        _ -> Err ("'*' expected a list of integers, got "
                  ++ I.valueToString args)
  in muler a |> Result.map I.Integer


div : Environment -> Value -> Result String Value
div e args = case args of
  I.Cons (I.Integer a) (I.Cons (I.Integer b) _) -> Ok (I.Integer (a // b))
  I.Cons (I.Integer a) (I.Integer b) -> Ok (I.Integer (a // b))
  _ -> Err ("'/' expected two integers, got "
            ++ I.valueToString args)


mod : Environment -> Value -> Result String Value
mod e args = case args of
  I.Cons (I.Integer a) (I.Cons (I.Integer b) _) -> Ok (I.Integer (modBy b a))
  I.Cons (I.Integer a) (I.Integer b) -> Ok (I.Integer (modBy b a))
  _ -> Err ("'%' expected two integers, got "
            ++ I.valueToString args)


gt : Environment -> Value -> Result String Value
gt e args = case args of
  I.Cons (I.Integer a) (I.Cons (I.Integer b) _) ->
    if a > b then Ok (I.Integer a)
    else Ok I.Null
  I.Cons (I.Integer a) (I.Integer b) ->
    if a > b then Ok (I.Integer a)
    else Ok I.Null
  _ -> Err ("'>' expected two integers, got "
            ++ I.valueToString args)


eq : Environment -> Value -> Result String Value
eq e args = case args of
  I.Cons a (I.Cons b _) ->
    if a == b then Ok a
    else Ok I.Null
  _ -> Err ("'=' expected two arguments, got "
            ++ I.valueToString args)


lambda : Environment -> Value -> Result String Value
lambda env args =
  let
      skeletonize params = case params of
        I.Null -> Ok I.Wildcard
        I.Symbol s -> Ok (I.Capture s)
        I.Cons car cdr -> case skeletonize car of
          Err s -> Err s
          Ok carPattern -> case skeletonize cdr of
            Err s -> Err s
            Ok cdrPattern -> Ok (I.ConsPattern carPattern cdrPattern)
        _ -> Err ("Argument patterns must consist of symbols, conses and Nulls,"
                  ++ " instead got " ++ I.valueToString params)
  in case args of
    I.Cons params (I.Cons body _) ->
      skeletonize params
      |> Result.map (\x -> I.Fun (I.Closure env x body))
    _ -> Err ("lambda requires an argument pattern and body, instead got"
              ++ I.valueToString args)


cons : Environment -> Value -> Result String Value
cons env args = case args of
  I.Cons a (I.Cons b _) -> Ok (I.Cons a b)
  _ -> Err ("'cons' expected two arguments, got "
            ++ I.valueToString args)


carOp : Environment -> Value -> Result String Value
carOp env args = case args of
  I.Cons (I.Cons x _) _ -> Ok x
  _ -> Err ("'car' expected a cons cell, got "
            ++ I.valueToString args)


cdrOp : Environment -> Value -> Result String Value
cdrOp env args = case args of
  I.Cons (I.Cons _ x) _ -> Ok x
  _ -> Err ("'cdr' expected a cons cell, got "
            ++ I.valueToString args)


quote : Environment -> Value -> Result String Value
quote env args = case args of
  I.Cons x _ -> Ok (I.Quoted x)
  _ -> Err ("'quote' expected an argument, got "
            ++ I.valueToString args)


eval : Environment -> Value -> Result String Value
eval env args = case args of
  I.Cons x _ -> I.eval env x
  _ -> Err ("'eval' expected an argument, got "
            ++ I.valueToString args)


apply : Environment -> Value -> Result String Value
apply env args = case args of
  I.Cons fun params -> I.call env fun params
  _ -> Err ("'apply' expected arguments, got "
            ++ I.valueToString args)


ifNull : Environment -> Value -> Result String Value
ifNull env args = case args of
  I.Cons cond (I.Cons ifTrue (I.Cons ifFalse _)) ->
    if cond == I.Null then I.eval env ifTrue
    else I.eval env ifFalse
  _ -> Err ("'if-null' expected three, got "
            ++ I.valueToString args)


ifInteger : Environment -> Value -> Result String Value
ifInteger env args = case args of
  I.Cons cond (I.Cons ifTrue (I.Cons ifFalse _)) -> case cond of
    I.Integer _ -> I.eval env ifTrue
    _ -> I.eval env ifFalse
  _ -> Err ("'if-integer' expected three, got "
            ++ I.valueToString args)


ifSymbol : Environment -> Value -> Result String Value
ifSymbol env args = case args of
  I.Cons cond (I.Cons ifTrue (I.Cons ifFalse _)) -> case cond of
    I.Symbol _ -> I.eval env ifTrue
    _ -> I.eval env ifFalse
  _ -> Err ("'if-symbol' expected three, got "
            ++ I.valueToString args)


ifQuoted : Environment -> Value -> Result String Value
ifQuoted env args = case args of
  I.Cons cond (I.Cons ifTrue (I.Cons ifFalse _)) -> case cond of
    I.Quoted _ -> I.eval env ifTrue
    _ -> I.eval env ifFalse
  _ -> Err ("'if-quoted' expected three, got "
            ++ I.valueToString args)


ifCons : Environment -> Value -> Result String Value
ifCons env args = case args of
  I.Cons cond (I.Cons ifTrue (I.Cons ifFalse _)) -> case cond of
    I.Cons _ _ -> I.eval env ifTrue
    _ -> I.eval env ifFalse
  _ -> Err ("'if-cons' expected three, got "
            ++ I.valueToString args)


ifFun : Environment -> Value -> Result String Value
ifFun env args = case args of
  I.Cons cond (I.Cons ifTrue (I.Cons ifFalse _)) -> case cond of
    I.Fun _ -> I.eval env ifTrue
    _ -> I.eval env ifFalse
  _ -> Err ("'if-function' expected three, got "
            ++ I.valueToString args)


stdDefs : Environment
stdDefs = Dict.fromList
  [ ("+", I.Fun (I.Builtin add))
  , ("-", I.Fun (I.Builtin sub))
  , ("*", I.Fun (I.Builtin mul))
  , ("/", I.Fun (I.Builtin div))
  , ("%", I.Fun (I.Builtin mod))
  , (">", I.Fun (I.Builtin gt))
  , ("=", I.Fun (I.Builtin eq))
  , (".", I.Fun (I.Builtin cons))
  , ("car", I.Fun (I.Builtin carOp))
  , ("cdr", I.Fun (I.Builtin cdrOp))
  , ("quote", I.Fun (I.Builtin quote))
  , ("lambda", I.Fun (I.Builtin lambda))
  , ("eval", I.Fun (I.Builtin eval))
  , ("apply", I.Fun (I.Builtin apply))
  , ("if-symbol", I.Fun (I.Builtin ifSymbol))
  , ("if-integer", I.Fun (I.Builtin ifInteger))
  , ("if-quote", I.Fun (I.Builtin ifQuoted))
  , ("if-cons", I.Fun (I.Builtin ifCons))
  , ("if-function", I.Fun (I.Builtin ifFun))
  , ("if-null", I.Fun (I.Builtin ifNull))
  ]


stdMacros : Environment
stdMacros = Dict.empty
